/* ----------------------------------------------------------------------
   LAMMPS - Large-scale Atomic/Molecular Massively Parallel Simulator
   http://lammps.sandia.gov, Sandia National Laboratories
   Steve Plimpton, sjplimp@sandia.gov

   Copyright (2003) Sandia Corporation.  Under the terms of Contract
   DE-AC04-94AL85000 with Sandia Corporation, the U.S. Government retains
   certain rights in this software.  This software is distributed under 
   the GNU General Public License.

   See the README file in the top-level LAMMPS directory.
------------------------------------------------------------------------- */

//#include "math.h"
#include "sort.h"
//#include "error.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
//#include "memory.h"

//using namespace LAMMPS_NS;

#define DUFF_DEVICE_4(count_,src,dest) \
{ \
	int times_ = (count_ + 3) >> 2; \
	double *src_ = src, *dest_ = dest; \
	switch (count_ & 3) { \
		case 0: do { *dest_++ = *src_++; \
		case 3: *dest_++ = *src_++; \
		case 2: *dest_++ = *src_++; \
		case 1: *dest_++ = *src_++; \
		} while (--times_ > 0); \
	} \
}

void heapsort(double buf[], unsigned int n, unsigned int size_one)
{
	unsigned int i = n/2, parent, child;
	int t;
	//double *tmp_buf =  (double *) memory->smalloc(n*size_one*sizeof(double),"sort:tmp_buf");
  double *tmp_buf = (double*) malloc(n*size_one*sizeof(double));

  for (;;) { /* Loops until buf is sorted */
		if (i > 0) { /* First stage - Sorting the heap */
			i--;           /* Save its index to i */
			t = (int)buf[i*size_one];    /* Save parent value to t */
			DUFF_DEVICE_4(size_one,&buf[i*size_one],tmp_buf);
		} else {     /* Second stage - Extracting elements in-place */
			n--;           /* Make the new heap smaller */
			if (n == 0) return; /* When the heap is empty, we are done */
			t = (int)buf[n*size_one];    /* Save last value (it will be overwritten) */
			DUFF_DEVICE_4(size_one,&buf[n*size_one],tmp_buf);
			DUFF_DEVICE_4(size_one,&buf[0],&buf[n*size_one]);
		}

		parent = i; /* We will start pushing down t from parent */
		child = i*2 + 1; /* parent's left child */

		 /* Sift operation - pushing the value of t down the heap */
		while (child < n) {
		  if (child + 1 < n  &&  (int)buf[(child + 1)*size_one] > (int)buf[child*size_one]) {
				child++; /* Choose the largest child */
			}
			if ((int)buf[child*size_one] > t) { /* If any child is bigger than the parent */
				DUFF_DEVICE_4(size_one,&buf[child*size_one],&buf[parent*size_one]); /* Move the largest child up */
				parent = child; /* Move parent pointer to this child */
				child = parent*2 + 1; /* Find the next child */
			} else {
				break; /* t's place is found */
			}
		}
		//buf[parent] = t; /* We save t in the heap */
		DUFF_DEVICE_4(size_one,tmp_buf,&buf[parent*size_one]);
	}
	//memory->sfree(tmp_buf);
	free(tmp_buf);
}

int main(int argc, char* argv[])
{
	int i, j;
	int length = 25;
	int array[] = {14, 0, 18, 17, 1, 12, 11, 21, 15, 10, 4, 8, 22, 20, 13, 5, 7, 6, 24, 16, 3, 2, 23, 9, 19};
	int size_one = 5;
	double *buf = (double*) malloc(length*size_one*sizeof(double));
	memset(buf,0,length*size_one*sizeof(double));

	for (i=0; i<length; i++) {
		buf[i*size_one] = array[i];
		buf[i*size_one+size_one-1] = i;
	}

	for (i=0; i<length; i++) {
		for (j=0; j<size_one; j++)
			printf("%1.1f ",buf[i*size_one+j]);
		printf("\n");
	}
	printf("\n");

	heapsort(buf, length, size_one);
	for (i=0; i<length; i++) {
		for (j=0; j<size_one; j++)
			printf("%1.1f ",buf[i*size_one+j]);
		printf("\n");
	}
	printf("\n");
	free(buf);
}
